#ifndef HWACCESS_INCLUDED
#define HWACCESS_INCLUDED



#ifndef TYPES_INCLUDED
    #error "Needs types(Types.inc)."
    //#include "Types.inc"
#endif

#ifndef GLOBALS_INCLUDED
    #error "Needs global types and variables(Globals.inc)."
    //#include "Globals.inc"
#endif

#ifndef HWCORE_INCLUDED
    #error "Needs hardware core(HWCore.inc)."
    //#include "HWCore.inc"
#endif

#ifndef ROBDEBUG_INCLUDED
    #error "Needs robust debugging(RobDebug.inc)."
    //#include "RobDebug.inc"
#endif

void PrinterPortByteWrite(U8 Value)
{   PDL("PrintPortByteWrite(U8 Value): Control Printer Port Digital Outputs");
    Deb("Do Printer Pins Default to output?");
    outportb(PRINTER_PORT_ADDRESS, Value);
}

void PCIDetect(U16 TargetVID, U16 TargetDID)
{
    U16 Bus, Device;
    union
    {
        struct
        {
            unsigned char Addr;
            unsigned char Function: 3;
            unsigned char Device: 5;
            unsigned char Bus;
            unsigned char Header;
        } P;
        unsigned long Trans;
    } Trans;

    DIO96Base[0] = 0;
    DIO96Base[1] = 0;

    Trans.P.Header = 0x80;
    Trans.P.Addr = 0;
    Trans.P.Function = 0;
    for ( Bus = 0; Bus < 256; ++Bus )
    {
        Trans.P.Bus = Bus;
        U16 DevicesOnBus = 0;
        for ( Device = 0; Device < 32; ++Device )
        {
            Trans.P.Device = Device;
            U16 ID;
            OutPortDW(0xCF8, Trans.Trans);
            ID = InPort(0xCFC);
            if ( ID == 0xFFFF ) continue;
            ++DevicesOnBus;
            if ( ID != TargetVID ) continue;
            ID = InPort(0xCFE);
            if ( ID != TargetDID ) continue;

            OutPortDW(0xCF8, Trans.Trans | 0x18);
            if ( DIO96Base[0] )
            {
                DIO96Base[1] = InPortDW(0xCFC) & 0xFFF8;
                break;
            }
            else
            {
                DIO96Base[0] = InPortDW(0xCFC) & 0xFFF8;
            }
        }
        if ( DIO96Base[1] ) break;
        if ( DevicesOnBus == 0 ) break;
    }

}

TCPUTime __fastcall GetTick(void)
{
    __emit__(0x0F, 0x31);
    asm SHRD EAX, EDX, 5;
    return _EAX;
}

void InitClockHz(void)
{
//   delay(1);
    TCPUTime Tick = GetTick();
    delay(1000);
    ClockHz = GetTick();
    ClockHz -= Tick;
    ClockHz *= 1;//0;
    USecsPerTick = 1000000.0L / double(ClockHz);
    MinsPerBigTick = 4294967296.0L * 32.0L / (60.0L * double(ClockHz));
}

U32 GetBigTick(void)
{
    __emit__(0x0F, 0x31);
    return _EDX;
}

void GetMasterTick(TMasterTime & Target)
{
    __emit__(0x0F, 0x31);
    Target.Lo32 = _EAX;
    Target.Hi32 = _EDX;
}

int __fastcall GotFrontData(void)
{
    return
        (((InPortB(DIO96Base[0] + 7) & 0x80) != 0) ? 1 : 0)
    |
        (((InPortB(DIO96Base[1] + 7) & 0x80) != 0 ) ? 2 : 0)
    ;
}

int __fastcall GotOtherData(void)
{
    return ((InPortB(DIO96Base[0] + 7) & 0x80) != 0) ? 1 : 0;
}

void __fastcall GetFrontDatum(void)                  // Take Data.  Must be FAST.  Abuses (makes optimum use of) structures to pump 32-bit data around fast.
{
    PDL("void GetFrontDatum(void): Gets one data point off the cards for two X I-Beams and one Y I-Beam.");
    pU32 pData = pU32(&MostRecentData.Datum.X[0]);
    *(pData++) = InPortDW(DIO96Base[0]);
    *(pData++) = InPortDW(DIO96Base[0]);
    *(pData++) = InPortDW(DIO96Base[0]);
    *(pData++) = InPortDW(DIO96Base[1]);
    *(pData  ) = InPortDW(DIO96Base[1]);
    InPortDW(DIO96Base[1]);
    MostRecentData.Datum.SampleTime = GetTick();
}

void __fastcall GetOtherDatum(void)                  // Take Data.  Must be FAST.  Abuses (makes optimum use of) structures to pump 32-bit data around fast.
{
    PDL("void GetOtherDatum(void): Gets one data point off the card for one X I-Beams and one Y I-Beam.");
    pU32 pData = pU32(&MostRecentData.Datum.X[0]);
    *(pData++) = InPortDW(DIO96Base[0]);
    *(pData++) = InPortDW(DIO96Base[0]);
    *(pData  ) = InPortDW(DIO96Base[0]);
    MostRecentData.Datum.SampleTime = GetTick();

}

void SetMask(void)
{
//    pU32 pData = Mask.By32;
    OutPortDW(DIO96Base[0] + 0x10, Mask.By32[0]);// *(pData++));
    OutPortDW(DIO96Base[0] + 0x14, Mask.By32[1]);// *(pData++));
    OutPortDW(DIO96Base[0] + 0x18, Mask.By32[2]);// *(pData++));
    if ( DIO96Base[1] )
    {
        OutPortDW(DIO96Base[1] + 0x10, Mask.By32[3]);//  *(pData++));
        OutPortDW(DIO96Base[1] + 0x14, Mask.By32[4]);//  *(pData  ));
    }
}

unsigned long BenchTime;

void __fastcall BenchOn(void)
{
  BenchTime = GetTick();
}

void __fastcall BenchOff(TBool Wait)
{
  register U32 BenchDone = GetTick();
  BenchDone -= BenchTime;
  double dT = BenchDone;
  dT *= USecsPerTick;
  char msg[80];
  sprintf(msg,"Benchmark: %lfus", dT);
    puts("");
    puts(msg);
  if (Wait)
    getch();
}

void __fastcall BenchOff(char* msg, TBool Wait)
{
  BenchOff(Wait);
  puts(msg);
}




#endif