#ifndef ROBDEBUG_INCLUDED
#define ROBDEBUG_INCLUDED



#ifndef TYPES_INCLUDED
    #error "Needs types(Types.inc)."
    //#include "Types.inc"
#endif

#ifndef HWCORE_INCLUDED
    #error "Needs hardware core(HWCore.inc)."
    //#include "HWCore.inc"
#endif

#define MAX_REMARK_SIZE (80*5-2)    // five lines of Debug info, but don't extra-wrap /r/n on last line kinda --J2H test fenceposts

#define VERBOSITY_S_SILENT 0        // 0.03 - J2H
#define VERBOSITY_S_TBD 10
#define VERBOSITY_S_DEB 20
#define VERBOSITY_S_REM 30
#define VERBOSITY_S_PDL 40
#define VERBOSITY_S_END 50

#define VERBOSITY_H_SILENT 0
#define VERBOSITY_H_DEBUGREPORT 10
#define VERBOSITY_H_DEBUGIO 20

#define SOURCE_VERBOSITY VERBOSITY_S_DEB
#define HARDWARE_VERBOSITY VERBOSITY_H_SILENT

#if SOURCE_VERBOSITY >= VERBOSITY_S_END
    #define USES_REMARK True
    #define End(x) Remark("END ", x)
#else
    #define End(x)
#endif

#if SOURCE_VERBOSITY >= VERBOSITY_S_PDL
    #define USES_REMARK True
    #define PDL(x) Remark("PDL ", x)
#else
    #define PDL(x)
#endif

#if SOURCE_VERBOSITY >= VERBOSITY_S_REM
    #define USES_REMARK True
    #define Rem(x) Remark("REM ", x)
#else
    #define Rem(x)
#endif

#if SOURCE_VERBOSITY >= VERBOSITY_S_DEB
    #define USES_REMARK True
    #define Deb(x) Remark("DEB ", x)
#else
    #define Deb(x)
#endif

#if SOURCE_VERBOSITY >= VERBOSITY_S_TBD
    #define USES_REMARK True
    #define TBD(x) Remark("TBD ", x)
#else
    #define TBD(x)
#endif

#ifndef USES_REMARK
    #define USES_REMARK False
#endif

#if USES_REMARK //if any Debug reporting enabled we need the Remark function, otherwise tighten the code.
    char *Remark(char *Prefix, char *Message)
    {
        //fputs("PDL char *Remark(char *Prefix, char *Buf): copy to stderr -and return- Prefix+Buf+CRLF.  For Nesting. Note:hard to print your own source code", stderr);

        static char PrintBuf[MAX_REMARK_SIZE];
        U16 BufLeft = sizeof(PrintBuf) - 2; //Account for CRLF.
        strncpy(PrintBuf, Prefix, BufLeft);
        BufLeft -= strlen(Prefix); //Account for prefix.
        strncat(PrintBuf, Message, BufLeft);
        strcat(PrintBuf, "\r\n");

        fputs(PrintBuf, stderr);
        return PrintBuf;
    }
#endif

//#define Remark(_x_, _y_)

#if HARDWARE_VERBOSITY >= VERBOSITY_H_DEBUGREPORT
    #define HDR(x) Remark("HDR ", x)
#else
    #define HDR(x)
#endif

#if HARDWARE_VERBOSITY >= VERBOSITY_H_DEBUGIO
    #define InPortB(x) DebugInPortB(x)
    #define InPort(x) DebugInPort(x)
    #define InPortDW(x) DebugInPortDW(x)
    #define INSW(x, y, z) DebugINSW(x, y, z)
    #define INSDW(x, y, z) DebugINSDW(x, y, z)
    #define OutPortB(x, y) DebugOutPortB(x, y)
    #define OutPort(x, y) DebugOutPort(x, y)
    #define OutPortDW(x, y) DebugOutPortDW(x, y)
#else
    #define InPortB(x) inportb(x)
    #define InPort(x) inport(x)
    //#define InPortDW(x) InPortDW(x)
    #define INSW(x, y, z) insw(x, y, z)
    #define INSDW(x, y, z) insdw(x, y, z)
    #define OutPortB(x, y) outportb(x, y)
    #define OutPort(x, y) outport(x, y)
    //#define OutPortDW(x, y) OutPortDW(x, y)
#endif

#define For(_params_) { for _params_ {
#define ForInt(_var_, _loval_, _hival_) { for (int _var_ = _loval_; _var_ <= _hival_; ++_var_) {
#define ForIntDown(_var_, _hival_, _loval_) { for (int _var_ = _hival_; _var_ >= _loval_; --_var_) {
#define EndFor } } End("FOR")

char *Print(char *Buf)
{
    //PDL("char *Print(char *Buf): copy Buf to stdout and return pointer to Buf (for nesting)");
    fputs(Buf,stdout);
    return Buf;
}



#endif